import requests
from .errors import ShareplumRequestError


def get(session, url, **kwargs):
    try:
        response = session.get(url, **kwargs)
        response.raise_for_status()
        return response
    except requests.exceptions.RequestException as err:
        raise ShareplumRequestError("Shareplum HTTP Get Failed", err)


def post(session, url, **kwargs):
    try:
        response = session.post(url, **kwargs)
        response.raise_for_status()
        return response
    except requests.exceptions.RequestException as err:
        raise ShareplumRequestError("Shareplum HTTP Post Failed", err)



import re
from datetime import datetime
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from .request_helper import post
import requests
import json
from lxml import etree

from .soap import Soap

# import defusedxml.ElementTree as etree


class _List2007:
    """Sharepoint Lists Web Service
       Microsoft Developer Network:
       The Lists Web service provides methods for working
       with SharePoint lists, content types, list items, and files.
    """

    def __init__(
        self,
        session,  # type: requests.Session
        list_name,  # type: str
        url,  # type: Callable[[str], str]
        verify_ssl,  # type: bool
        users,  # type: Optional[Dict]
        huge_tree,  # type: bool
        timeout,  # type: Optional[int]
        exclude_hidden_fields=False,  # type: bool
        site_url=None,
    ):
        # type: (...) -> None
        self._session = session
        self.list_name = list_name
        self._url = url
        self._verify_ssl = verify_ssl
        self.users = users
        self.huge_tree = huge_tree
        self.timeout = timeout
        self._exclude_hidden_fields = exclude_hidden_fields
        # List Info
        self.fields = []  # type: List[Dict[str, str]]
        self.regional_settings = {}  # type: Dict[str, str]
        self.server_settings = {}  # type: Dict[str, str]
        self.get_list()
        self.views = self.get_view_collection()
        self.version = "2007"

        # fields sometimes share the same displayname
        # filtering fields to only contain visible fields
        # minimizes the chance of a one field hiding another
        if exclude_hidden_fields:
            self.fields = [field for field in self.fields if field.get("Hidden", "FALSE") == "FALSE"]

        self._sp_cols = {i["Name"]: {"name": i["DisplayName"], "type": i["Type"]} for i in self.fields}
        self._disp_cols = {i["DisplayName"]: {"name": i["Name"], "type": i["Type"]} for i in self.fields}

        title_col = self._sp_cols["Title"]["name"]
        title_type = self._sp_cols["Title"]["type"]
        self._disp_cols[title_col] = {"name": "Title", "type": title_type}
        self.last_request = None  # type: Optional[str]
        self.date_format = re.compile("[0-9]+-[0-9]+-[0-9]+ [0-9]+:[0-9]+:[0-9]+")

    def _headers(self, soapaction):
        # type: (str) -> Dict[str,str]
        headers = {
            "Content-Type": "text/xml; charset=UTF-8",
            "SOAPAction": "http://schemas.microsoft.com/sharepoint/soap/" + soapaction,
        }
        return headers

    def _mutate_to_internal(self, data):
        # type: (List[Dict]) -> None
        """From 'Column Title' to 'Column_x0020_Title'"""
        for _dict in data:
            keys = list(_dict.keys())[:]
            for key in keys:
                if key not in self._disp_cols:
                    raise Exception(key + " not a column in current List.")
                _dict[self._disp_cols[key]["name"]] = self._sp_type(key, _dict.pop(key))

    def _convert_to_internal(self, data):
        # type: (List[Dict]) -> None
        """From 'Column Title' to 'Column_x0020_Title'"""
        new_data = []
        for _dict in data:
            keys = list(_dict.keys())[:]
            new_dict = dict()
            for key in keys:
                if key not in self._disp_cols:
                    raise Exception(key + " not a column in current List.")
                new_dict[self._disp_cols[key]["name"]] = self._sp_type(key, _dict[key])
            new_data.append(new_dict)

        return new_data

    def _convert_to_display(self, data):
        # type: (List[Dict]) -> None
        """From 'Column_x0020_Title' to  'Column Title'"""
        for _dict in data:
            keys = list(_dict.keys())[:]
            for key in keys:
                if key not in self._sp_cols:
                    raise Exception(key + " not a column in current List.")
                _dict[self._sp_cols[key]["name"]] = self._python_type(key, _dict.pop(key))

    def _python_type(self, key, value):
        # type: (str, Any) -> Any
        """Returns proper type from the schema"""
        try:
            field_type = self._sp_cols[key]["type"]
            if field_type in ["Number", "Currency"]:
                return float(value)
            elif field_type == "DateTime":

                # Need to remove the '123;#' from created dates, but we will do it for all dates
                # self.date_format = re.compile('\d+-\d+-\d+ \d+:\d+:\d+')
                match = self.date_format.search(value)
                if match:
                    value = match.group(0)

                # NOTE: I used to round this just date (7/28/2018)
                return datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
            elif field_type == "Boolean":
                if value == "1":
                    return "Yes"
                elif value == "0":
                    return "No"
                else:
                    return ""
            elif field_type in ("User", "UserMulti"):
                # Sometimes the User no longer exists or
                # has a diffrent ID number so we just remove the "123;#"
                # from the beginning of their name
                if self.users and value in self.users["sp"]:
                    return self.users["sp"][value]
                elif "#" in value:
                    users = []
                    for i, value in enumerate(value.split(';#')):
                        if i % 2 == 0:
                            user = '#%s' % value
                        else:
                            user += ';#%s' % value
                            users.append(user)
                    return users
                else:
                    return value
            else:
                return value
        except AttributeError:
            # TODO: log me
            return value

    def _sp_type(self, key, value):
        # type: (str, Any) -> Any
        """Returns proper type from the schema"""
        try:
            field_type = self._disp_cols[key]["type"]
            if field_type in ["Number", "Currency"]:
                return value
            elif field_type == "DateTime":
                return value.strftime("%Y-%m-%d %H:%M:%S")
            elif field_type == "Boolean":
                if value == "Yes":
                    return "1"
                elif value == "No":
                    return "0"
                else:
                    raise Exception("%s not a valid Boolean Value, only 'Yes' or 'No'" % value)
            elif self.users and field_type == "User":
                return self.users["py"][key]
            else:
                return value
        except AttributeError:
            # TODO: Log me
            return value

    def get_list_items(
        self,
        view_name=None,  # type: Optional[str]
        fields=None,  # type: Optional[List[str]]
        query=None,  # type: Optional[Dict]
        row_limit=0,  # type: int
        debug=False,  # type: bool
    ):
        # type: (...) -> Optional[Any]
        """Get Items from current list
           row_limit defaulted to 0 (unlimited)
        """

        # Build Request
        soap_request = Soap("GetListItems")
        soap_request.add_parameter("listName", self.list_name)
        # Convert Displayed View Name to View ID
        if view_name:
            soap_request.add_parameter("viewName", self.views[view_name]["Name"][1:-1])

        # Add viewFields
        if fields:
            # Convert to SharePoint Style Column Names
            for i, val in enumerate(fields):
                fields[i] = self._disp_cols[val]["name"]
            viewfields = fields
            soap_request.add_view_fields(fields)
            # Check for viewname and query
            if [view_name, query] == [None, None]:
                # Add a query if the viewname and query are not provided
                # We sort by 'ID' here Ascending is the default
                soap_request.add_query({"OrderBy": ["ID"]})

        elif view_name:
            viewfields = self.get_view(view_name)["fields"]
        else:
            # No fields or views provided so get everything
            viewfields = [x for x in self._sp_cols]

        # Add query
        if query:
            modified_query = dict()
            where = etree.Element('Where')

            parents = [where]
            if "Where" in query:
                where = etree.Element("Where")

                parents = [where]
                for _i, field in enumerate(query["Where"]):
                    if field == "And":
                        parents.append(etree.SubElement(parents[-1], "And"))
                    elif field == "Or":
                        if parents[-1].tag == "Or":
                            parents.pop()
                        parents.append(etree.SubElement(parents[-1], "Or"))
                    else:
                        _type = etree.SubElement(parents[-1], field[0])
                        field_ref = etree.SubElement(_type, "FieldRef")
                        field_ref.set("Name", self._disp_cols[field[1]]["name"])
                        value = etree.SubElement(_type, "Value")
                        value.set("Type", self._disp_cols[field[1]]["type"])
                        value.text = self._sp_type(field[1], field[2])

                # query["Where"] = where
                modified_query["Where"] = where

            soap_request.add_query(modified_query)

        # Set Row Limit
        soap_request.add_parameter("rowLimit", str(row_limit))
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Lists"),
                        headers=self._headers("GetListItems"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        # Parse Response
        # TODO: Verify if this works with Sharepoint lists with validation
        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        listitems = envelope[0][0][0][0][0]
        data = []
        for row in listitems:
            # Strip the 'ows_' from the beginning with key[4:]
            data.append({key[4:]: value for (key, value) in row.items() if key[4:] in viewfields})

        self._convert_to_display(data)

        if debug:
            return response
        else:
            return data

    def get_list(self):  # type: () -> None
        """Get Info on Current List
           This is run in __init__ so you don't
           have to run it again.
           Access from self.schema
        """

        # Build Request
        soap_request = Soap("GetList")
        soap_request.add_parameter("listName", self.list_name)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Lists"),
                        headers=self._headers("GetList"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        # Parse Response
        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))  # type: etree.ElementTree
        (fields, regional_settings, server_settings) = self.parse_list_envelope(envelope)
        self.fields += fields
        self.regional_settings.update(regional_settings)
        self.server_settings.update(server_settings)

    @staticmethod
    def parse_list_envelope(envelope):
        # type: (etree.ElementTree) -> Tuple[List[Dict[str, Any]], Dict[str, str], Dict[str, str]]
        _list = envelope[0][0][0][0]
        fields = []
        regional_settings = dict()
        server_settings = dict()

        # info = {key: value for (key, value) in _list.items()}
        for row in _list.xpath(
            "//*[re:test(local-name(), '.*Fields.*')]", namespaces={"re": "http://exslt.org/regular-expressions"}
        )[0].getchildren():
            fields.append({key: value for (key, value) in row.items()})

        for setting in _list.xpath(
            "//*[re:test(local-name(), '.*RegionalSettings.*')]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )[0].getchildren():
            regional_settings[setting.tag.strip("{http://schemas.microsoft.com/sharepoint/soap/}")] = setting.text

        for setting in _list.xpath(
            "//*[re:test(local-name(), '.*ServerSettings.*')]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )[0].getchildren():
            server_settings[setting.tag.strip("{http://schemas.microsoft.com/sharepoint/soap/}")] = setting.text

        return fields, regional_settings, server_settings

    def get_view(self, view_name):  # type: (str)  -> Optional[Dict]
        """Get Info on View Name
        """

        # Build Request
        soap_request = Soap("GetView")
        soap_request.add_parameter("listName", self.list_name)

        if not view_name:
            views = self.get_view_collection()
            if views:
                for v in views:
                    if "DefaultView" in v:
                        if views[v]["DefaultView"] == "TRUE":
                            view_name = v
                            break

        if self.list_name not in ["UserInfo", "User Information List"] and self.views:
            soap_request.add_parameter("viewName", self.views[view_name]["Name"][1:-1])
        else:
            soap_request.add_parameter("viewName", view_name)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Views"),
                        headers=self._headers("GetView"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        # Parse Response
        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))  # type: etree.ElementTree
        # TODO: Fix me? Should this use XPath too?
        view = envelope[0][0][0][0]
        info = {key: value for (key, value) in view.items()}
        fields = [x.items()[0][1] for x in view[1]]
        return {"info": info, "fields": fields}

    def get_view_collection(self):  # type: () -> Optional[Dict[str, Dict[str, str]]]
        """Get Views for Current List
           This is run in __init__ so you don't
           have to run it again.
           Access from self.views
        """

        # Build Request
        soap_request = Soap("GetViewCollection")
        soap_request.add_parameter("listName", self.list_name)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Views"),
                        headers=self._headers("GetViewCollection"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        views = envelope[0][0][0][0]
        data = []
        for row in views.getchildren():
            data.append({key: value for (key, value) in row.items()})
        view = {}
        for row in data:
            view[row["DisplayName"]] = row
        return view

    def get_version_collection(self, list_id, item_id, field_name):  # type: () -> List[Dict[str, str]]

        # Build Request
        soap_request = Soap("GetVersionCollection")
        soap_request.add_parameter("strlistID", list_id)
        soap_request.add_parameter("strlistItemID", item_id)
        soap_request.add_parameter("strFieldName", field_name)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Lists"),
                        headers=self._headers("GetVersionCollection"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        # fix invalid attribute name: Sharepoints uses the field name as it is,
        # including whitespaces and special characters, as attribute name
        # for the Version element. To enable successful parsing, we replace
        # the attribute name (which we know anyway) by a constant, e.g. field_name
        content = response.text
        content = content.replace("Version {field_name}=\"".format(field_name=field_name), "Version field_name=\"")

        envelope = etree.fromstring(content.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        versions = envelope[0][0][0][0]
        data = []
        for row in versions.getchildren():
            data.append({
                'content': row.attrib['field_name'],
                'modified': row.attrib['Modified'],
                'editor': row.attrib['Editor']})
        return data

    def update_list_items(self, data, kind, mutate_data=False):  # type: (List[Dict[str, str]], str) -> Any
        """Update List Items
           kind = 'New', 'Update', or 'Delete'

           New:
           Provide data like so:
               data = [{'Title': 'New Title', 'Col1': 'New Value'}]

           Update:
           Provide data like so:
               data = [{'ID': 23, 'Title': 'Updated Title'},
                       {'ID': 28, 'Col1': 'Updated Value'}]

           Delete:
           Just provided a list of ID's
               data = [23, 28]
        """
        if type(data) != list:
            raise Exception("data must be a list of dictionaries")
        # Build Request
        soap_request = Soap("UpdateListItems")
        soap_request.add_parameter("listName", self.list_name)
        if kind != "Delete":
            if mutate_data:
                spdata = data
                self._mutate_to_internal(spdata)
            else:
                spdata = self._convert_to_internal(data)
        else:
            spdata = data

        soap_request.add_actions(spdata, kind)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Lists"),
                        headers=self._headers("UpdateListItems"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        # Parse Response
        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        # TODO: Fix me
        results = envelope[0][0][0][0]
        data_out = {}  # type: Dict
        for result in results:
            if result.text != "0x00000000" and result[0].text != "0x00000000":
                data_out[result.attrib["ID"]] = (result[0].text, result[1].text)
            else:
                data_out[result.attrib["ID"]] = result[0].text
        return data_out

    def get_attachment_collection(self, _id):  # type: (str) -> Any
        """Get Attachments for given List Item ID"""

        # Build Request
        soap_request = Soap("GetAttachmentCollection")
        soap_request.add_parameter("listName", self.list_name)
        soap_request.add_parameter("listItemID", _id)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Lists"),
                        headers=self._headers("GetAttachmentCollection"),
                        data=str(soap_request).encode("utf-8"),
                        verify=False,
                        timeout=self.timeout)

        # Parse Request
        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        # TODO: Fix this
        attaches = envelope[0][0][0][0]
        attachments = []
        for attachment in attaches.getchildren():
            attachments.append(attachment.text)
        return attachments

    # Legacy API
    GetList = get_list
    GetListItems = get_list_items
    GetView = get_view
    GetViewCollection = get_view_collection
    GetAttachmentCollection = get_attachment_collection
    UpdateListItems = update_list_items


class _List365(_List2007):
    def __init__(self,
                 session,  # type: requests.Session
                 list_name,  # type: str
                 url,  # type: Callable[[str], str]
                 verify_ssl,  # type: bool
                 users,  # type: Optional[Dict]
                 huge_tree,  # type: bool
                 timeout,  # type: Optional[int]
                 exclude_hidden_fields=False,  # type: bool
                 site_url=None):
        super().__init__(session, list_name, url, verify_ssl, users, huge_tree, timeout, exclude_hidden_fields, site_url)
        self.site_url = site_url
        self.schema = self._get_schema()
        self.version = "v365"

    def _get_schema(self):
        url = self.site_url + f"/_api/lists/getbytitle('{self.list_name}')/RenderListDataAsStream"

        body = json.dumps({"parameters": {"RenderOptions": 4}})

        headers = {'Accept': 'application/json;odata=verbose',
                   'Content-Type': 'application/json;odata=verbose',
                   'X-RequestDigest': self.contextinfo['FormDigestValue']}

        response = post(self._session, url=url, headers=headers, data=body, timeout=self.timeout)
        return response.json()

    @property
    def contextinfo(self):
        response = post(self._session, self.site_url + "/_api/contextinfo")
        data = json.loads(response.text)
        return data

    @property
    def info(self):
        return self.GetList()

    # @property
    # def views(self):
    #     return self.GetViewCollection()

    def create_field(self, title, field_type=2, required="false", unique="false", static_name=None):
        update_data = {}
        update_data['__metadata'] = {'type': 'SP.Field'}
        update_data['Title'] = title
        update_data['FieldTypeKind'] = field_type
        update_data['Required'] = required
        update_data['EnforceUniqueValues'] = unique
        update_data['StaticName'] = static_name
        body = json.dumps(update_data)

        url = self.site_url + f"/_api/lists/getbytitle('{self.list_name}')/Fields"

        headers = {'Accept': 'application/json;odata=verbose',
                   'Content-Type': 'application/json;odata=verbose',
                   'X-RequestDigest': self.contextinfo['FormDigestValue']}

        response = post(self._session, url=url, headers=headers, data=body, timeout=self.timeout)
        return response.json()



from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import requests
from requests.packages.urllib3.util.retry import Retry
from requests_toolbelt import SSLAdapter
from lxml import etree
# import defusedxml.ElementTree as etree

from .request_helper import get, post
from .list import _List2007, _List365
from .folder import _Folder
from .soap import Soap
from .version import __version__

from enum import Enum


# SharePoint Versions
class Version(Enum):
    v2007 = 1
    v2010 = 2
    v2013 = 3
    v2016 = 4
    v2019 = 5
    v365 = 6


class _Site2007:

    def __init__(self,
                 site_url,  # type: str
                 auth=None,  # type: Optional[Any]
                 authcookie=None,  # type: Optional[requests.cookies.RequestsCookieJar]
                 verify_ssl=True,  # type: bool
                 ssl_version=None,  # type: Optional[float]
                 huge_tree=False,  # type: bool
                 timeout=None,  # type: Optional[int]
                 retry=None):
        self.site_url = site_url
        self._verify_ssl = verify_ssl

        if retry is None:
            retry = Retry(total=5,
                        read=5,
                        connect=5,
                        backoff_factor=0.3,
                        status_forcelist=[500, 502, 503, 504])

        http_adaptor = requests.adapters.HTTPAdapter(max_retries=retry)
        https_adaptor = http_adaptor

        self._session = requests.Session()
        if ssl_version is not None:
            https_adaptor = SSLAdapter(ssl_version, max_retries=retry)

        self._session.mount("https://", https_adaptor)
        self._session.mount("http://", http_adaptor)

        self._session.headers.update({"user-agent": "shareplum/%s" % __version__})

        if authcookie is not None:
            self._session.cookies = authcookie
        else:
            self._session.auth = auth

        self.huge_tree = huge_tree

        self.timeout = timeout

        self.last_request = None  # type: Optional[str]

        self._services_url = {
            "Alerts": "/_vti_bin/Alerts.asmx",
            "Authentication": "/_vti_bin/Authentication.asmx",
            "Copy": "/_vti_bin/Copy.asmx",
            "Dws": "/_vti_bin/Dws.asmx",
            "Forms": "/_vti_bin/Forms.asmx",
            "Imaging": "/_vti_bin/Imaging.asmx",
            "DspSts": "/_vti_bin/DspSts.asmx",
            "Lists": "/_vti_bin/lists.asmx",
            "Meetings": "/_vti_bin/Meetings.asmx",
            "People": "/_vti_bin/People.asmx",
            "Permissions": "/_vti_bin/Permissions.asmx",
            "SiteData": "/_vti_bin/SiteData.asmx",
            "Sites": "/_vti_bin/Sites.asmx",
            "Search": "/_vti_bin/Search.asmx",
            "UserGroup": "/_vti_bin/usergroup.asmx",
            "Versions": "/_vti_bin/Versions.asmx",
            "Views": "/_vti_bin/Views.asmx",
            "WebPartPages": "/_vti_bin/WebPartPages.asmx",
            "Webs": "/_vti_bin/Webs.asmx",
        }  # type: Dict[str, str]

        self.site_info = self.get_site()
        self.users = self.get_users()
        self.version = "2007"  # For Debugging

    def _url(self, service):
        # type: (str) -> str
        """Full SharePoint Service URL"""
        return "".join([self.site_url, self._services_url[service]])

    def _headers(self, soap_action):
        # type: (str) -> Dict[str, str]
        headers = {
            "Content-Type": "text/xml; charset=UTF-8",
            "SOAPAction": "http://schemas.microsoft.com/sharepoint/soap/" + soap_action,
        }
        return headers

    # This is part of List but seems awkward under the List Method
    def add_list(self, list_name, description, template_id):
        # type: (str, str, str) -> Any
        """Create a new List
           Provide: List Name, List Description, and List Template
           Templates Include:
               Announcements
               Contacts
               Custom List
               Custom List in Datasheet View
               DataSources
               Discussion Board
               Document Library
               Events
               Form Library
               Issues
               Links
               Picture Library
               Survey
               Tasks
        """
        template_ids = {
            "Announcements": "104",
            "Contacts": "105",
            "Custom List": "100",
            "Custom List in Datasheet View": "120",
            "DataSources": "110",
            "Discussion Board": "108",
            "Document Library": "101",
            "Events": "106",
            "Form Library": "115",
            "Issues": "1100",
            "Links": "103",
            "Picture Library": "109",
            "Survey": "102",
            "Tasks": "107",
        }

        # Let's automatically convert the different
        # ways we can select the template_id
        if type(template_id) == int:
            template_id = str(template_id)
        elif type(template_id) == str:
            if template_id.isdigit():
                pass
            else:
                template_id = template_ids[template_id]

        # Build Request
        soap_request = Soap("AddList")
        soap_request.add_parameter("listName", list_name)
        soap_request.add_parameter("description", description)
        soap_request.add_parameter("templateID", template_id)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Lists"),
                        headers=self._headers("AddList"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        return response.text

    def delete_list(self, list_name):
        # type: (str) -> Optional[str]
        """Delete a List with given name"""

        # Build Request
        soap_request = Soap("DeleteList")
        soap_request.add_parameter("listName", list_name)
        self.last_request = str(soap_request)

        # Send Request
        post(self._session,
             url=self._url("Lists"),
             headers=self._headers("DeleteList"),
             data=str(soap_request).encode("utf-8"),
             verify=self._verify_ssl,
             timeout=self.timeout)

    def get_form_collection(self, list_name):

        # Build Request
        soap_request = Soap("GetFormCollection")
        soap_request.add_parameter("listName", list_name)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Forms"),
                        headers=self._headers("GetFormCollection"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        items = envelope[0][0][0][0]
        data = []
        for _item in items:
            data.append({k: v for (k, v) in _item.items()})

        return data

    def get_site(self):

        # Build Request
        soap_request = Soap("GetSite")
        soap_request.add_parameter("SiteUrl", self.site_url)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Sites"),
                        headers=self._headers("GetSite"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        data = envelope[0][0][0]

        # TODO: Not sure what to do with this, so just return the text
        return data.text

    def get_list_templates(self):

        # Build Request
        soap_request = Soap("GetListTemplates")
        soap_request.add_parameter("GetListTemplates")
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Webs"),
                        headers=self._headers("GetListTemplates"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        lists = envelope[0][0][0][0]
        data = []
        for _list in lists:
            data.append({k: v for (k, v) in _list.items()})

        return data

    def get_site_templates(self, lcid="1033"):

        # Build Request
        soap_request = Soap("GetSiteTemplates")
        soap_request.add_parameter("LCID", lcid)
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Sites"),
                        headers=self._headers("GetSiteTemplates"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        return response
        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        lists = envelope[0][0][1]
        data = []
        for _list in lists:
            data.append({k: v for (k, v) in _list.items()})

        return data

    def get_list_collection(self):
        # type: () -> Optional[List[Dict[str, str]]]
        """Returns List information for current Site"""
        # Build Request
        soap_request = Soap("GetListCollection")
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("SiteData"),
                        headers=self._headers("GetListCollection"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        envelope = etree.fromstring(response.text.encode("utf-8"),
                                    parser=etree.XMLParser(huge_tree=self.huge_tree,
                                    recover=True))
        # TODO: Verify if this works on Sharepoint lists with validation
        lists = envelope[0][0][1]
        data = []
        for _list in lists:
            _list_data = {}
            for item in _list:
                key = item.tag.replace("{http://schemas.microsoft.com/sharepoint/soap/}", "")
                value = item.text
                _list_data[key] = value
            data.append(_list_data)

        return data

    def get_users(self, rowlimit=0):
        # type: (int) -> Optional[Dict[str, Dict[str, str]]]
        """Get Items from current list
           rowlimit defaulted to 0 (no limit)
        """

        # Build Request
        soap_request = Soap("GetListItems")
        soap_request.add_parameter("listName", "UserInfo")

        # Set Row Limit
        soap_request.add_parameter("rowLimit", str(rowlimit))
        self.last_request = str(soap_request)

        # Send Request
        response = post(self._session,
                        url=self._url("Lists"),
                        headers=self._headers("GetListItems"),
                        data=str(soap_request).encode("utf-8"),
                        verify=self._verify_ssl,
                        timeout=self.timeout)

        # Parse Response
        try:
            envelope = etree.fromstring(response.text.encode("utf-8"),
                                        parser=etree.XMLParser(huge_tree=self.huge_tree,
                                        recover=True))
        except Exception as e:
            raise requests.ConnectionError("GetUsers GetListItems response failed to parse correctly: " + str(e))
        # TODO: Verify if this works on Sharepoint lists with validation
        listitems = envelope[0][0][0][0][0]
        data = []
        for row in listitems:
            # Strip the 'ows_' from the beginning with key[4:]
            data.append({key[4:]: value for (key, value) in row.items() if key[4:]})

        return {
            "py": {i["ImnName"]: i["ID"] + ";#" + i["ImnName"] for i in data},
            "sp": {i["ID"] + ";#" + i["ImnName"]: i["ImnName"] for i in data},
        }

    # SharePoint Method Objects
    # Not the best name as it could clash with the built-in list()
    def list(self, list_name, exclude_hidden_fields=False):
        # type: (str, bool) -> _List2007
        """Sharepoint Lists Web Service
           Microsoft Developer Network:
           The Lists Web service provides methods for working
           with SharePoint lists, content types, list items, and files.
        """

        return _List2007(
            self._session,
            list_name,
            self._url,
            self._verify_ssl,
            self.users,
            self.huge_tree,
            self.timeout,
            exclude_hidden_fields=exclude_hidden_fields,
            site_url=self.site_url,
        )

    # Legacy API
    List = list
    GetUsers = get_users
    GetListCollection = get_list_collection
    DeleteList = delete_list
    AddList = add_list


class _Site365(_Site2007):
    def __init__(self,
                 site_url,  # type: str
                 auth=None,  # type: Optional[Any]
                 authcookie=None,  # type: Optional[requests.cookies.RequestsCookieJar]
                 verify_ssl=True,  # type: bool
                 ssl_version=None,  # type: Optional[float]
                 huge_tree=False,  # type: bool
                 timeout=None):  # type: Optional[int]
        super().__init__(site_url, auth, authcookie, verify_ssl, ssl_version, huge_tree, timeout)

        self._session.headers.update({'Accept': 'application/json',
                                      'Content-Type': 'application/json;odata=nometadata'})
        self.version = "v365"

    @property
    def info(self):
        response = get(self._session, self.site_url + "/_api/site")
        return response.json()

    def Folder(self, folder_name):
        """Sharepoint Folder Web Service
        """
        return _Folder(self._session, folder_name, self.site_url, timeout=self.timeout)

    def _get_form_digest_value(self):
        response = post(self._session, self.site_url + "/_api/contextinfo")
        return response.json()['FormDigestValue']

    @property
    def contextinfo(self):
        response = post(self._session, self.site_url + "/_api/contextinfo")
        return response.json()

    @property
    def contenttypes(self):
        response = get(self._session, self.site_url + "/_api/web/contenttypes")
        return response.json()['value']

    @property
    def eventreceivers(self):
        response = get(self._session, self.site_url + "/_api/web/eventreceivers")
        return response.json()['value']

    @property
    def features(self):
        response = get(self._session, self.site_url + "/_api/web/features")
        return response.json()['value']

    @property
    def fields(self):
        response = get(self._session, self.site_url + "/_api/web/fields")
        return response.json()['value']

    @property
    def lists(self):
        return self.GetListCollection()

    # This is a duplicate, but in REST
    # def GetListCollection(self):
    #     response = get(self._session, self.site_url + "/_api/web/lists")
    #     data = json.loads(response.text)
    #     return data['value']

    @property
    def siteusers(self):
        return self.GetUsers()

    def GetUsers(self, row_limit=None):
        response = get(self._session, self.site_url + "/_api/web/siteusers")
        return response.json()['value']

    @property
    def groups(self):
        response = get(self._session, self.site_url + "/_api/web/sitegroups")
        return response.json()['value']

    @property
    def roleassignments(self):
        response = get(self._session, self.site_url + "/_api/web/roleassignments")
        return response.json()['value']

    @property
    def web(self):
        response = get(self._session, self.site_url + "/_api/web")
        return response.json()['value']

    # SharePoint Method Objects
    # Not the best name as it could clash with the built-in list()
    def list(self, list_name, exclude_hidden_fields=False):
        # type: (str, bool) -> _List365
        """Sharepoint Lists Web Service
           Microsoft Developer Network:
           The Lists Web service provides methods for working
           with SharePoint lists, content types, list items, and files.
        """

        return _List365(
            self._session,
            list_name,
            self._url,
            self._verify_ssl,
            self.users,
            self.huge_tree,
            self.timeout,
            exclude_hidden_fields=exclude_hidden_fields,
            site_url=self.site_url,
        )

    # Legacy API
    List = list


def Site(site_url,  # type: str
         version=Version.v2007,
         auth=None,  # type: Optional[Any]
         authcookie=None,  # type: Optional[requests.cookies.RequestsCookieJar]
         verify_ssl=True,  # type: bool
         ssl_version=None,  # type: Optional[float]
         huge_tree=False,  # type: bool
         timeout=None):  # type: Optional[int]

    # We ask for the various versions of SharePoint with 2010 as default
    # Multiple Version are allowed, but only 2010, 2013, and 365 are implemented
    if version == Version.v2007:
        return _Site2007(site_url,
                         auth,
                         authcookie,
                         verify_ssl,
                         ssl_version,
                         huge_tree,
                         timeout)

    elif version == Version.v2010:
        return _Site2007(site_url,
                         auth,
                         authcookie,
                         verify_ssl,
                         ssl_version,
                         huge_tree,
                         timeout)

    elif version == Version.v2013:
        return _Site365(site_url,
                        auth,
                        authcookie,
                        verify_ssl,
                        ssl_version,
                        huge_tree,
                        timeout)

    elif version == Version.v2016:
        return _Site365(site_url,
                        auth,
                        authcookie,
                        verify_ssl,
                        ssl_version,
                        huge_tree,
                        timeout)

    elif version == Version.v2019:
        return _Site365(site_url,
                        auth,
                        authcookie,
                        verify_ssl,
                        ssl_version,
                        huge_tree,
                        timeout)

    elif version == Version.v365:
        return _Site365(site_url,
                        auth,
                        authcookie,
                        verify_ssl,
                        ssl_version,
                        huge_tree,
                        timeout)



from shareplumfreightplusio import Site
from shareplumfreightplusio import Office365
from shareplumfreightplusio.site import Version
from .test_settings import TEST_SETTINGS, TEST_PASSWORD
import unittest

# Edit test_server.json file to setup SharePoint Test Server
# Use OS Enviroment Variable TEST_PASSWORD for SharePoint password
# export TEST_PASSWORD=********


class SiteTestCase(unittest.TestCase):

    def setUp(self):
        if TEST_SETTINGS["version"] in ["2014", "2016", "2019", "365"]:
            version = Version.v2016
        else:
            version = Version.v2007

        authcookie = Office365(TEST_SETTINGS["server_url"], username=TEST_SETTINGS["username"], password=TEST_PASSWORD).GetCookies()
        self.site = Site(TEST_SETTINGS["site_url"], version=version, authcookie=authcookie)
        self.test_list = TEST_SETTINGS["test_list"]

    def tearDown(self):
        self.site._session.close()

    def test_site_roleassignments(self):
        print("Role Assignments")
        self.assertIsNotNone(self.site.roleassignments)

    def test_site_lists(self):
        print("Site.Lists")
        self.assertIsNotNone(self.site.lists)

    def test_site_users(self):
        print("Site.Users")
        self.assertIsNotNone(self.site.users)



import os
try:
    from .local_test_settings import TEST_SETTINGS
except ImportError:
    TEST_SETTINGS = {
        "server_url": "https://jrollins.sharepoint.com",
        "version": "2016",
        "site_url": "https://jrollins.sharepoint.com/sites/TestSite",
        "username": "jrollins@jrollins.onmicrosoft.com",
        "test_list": "BlahTestList123",
        "test_folder": "Shared Documents/This Folder",
        "test_folder_apostrophe": "Shared Documents/This' Folder"
    }
TEST_PASSWORD = os.environ.get('TEST_PASSWORD')
assert TEST_PASSWORD, "TEST_PASSWORD environment variable not set."



class ShareplumError(Exception):
    def __init__(self, msg, details=None):
        if details:
            super().__init__(f"{msg} : {details}")
        else:
            super().__init__(msg)


class ShareplumRequestError(ShareplumError):
    pass



import os
from lxml import etree
from shareplumfreightplusio.list import _List2007

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def test_xml_with_validation():
    with open(os.path.join(__location__, "data", "validation.xml")) as f:
        xml: str = f.read()
    envelope: etree.ElementTree = etree.fromstring(xml.encode("utf-8"), parser=etree.XMLParser(huge_tree=True, recover=True))
    fields, regional_settings, server_settings = _List2007.parse_list_envelope(envelope)
    print(fields)
    assert len(fields) > 0


def test_xml_sharepoint_2010():
    with open(os.path.join(__location__, "data", "2010xml.xml")) as f:
        xml: str = f.read()
    envelope: etree.ElementTree = etree.fromstring(xml.encode("utf-8"), parser=etree.XMLParser(huge_tree=True, recover=True))
    fields, regional_settings, server_settings = _List2007.parse_list_envelope(envelope)
    print(fields)
    assert len(fields) > 0



from shareplumfreightplusio import Site
from shareplumfreightplusio import Office365
from shareplumfreightplusio.site import Version
from .test_settings import TEST_SETTINGS, TEST_PASSWORD
import unittest

# Edit test_server.json file to setup SharePoint Test Server
# Use OS Enviroment Variable TEST_PASSWORD for SharePoint password
# export TEST_PASSWORD=********


class ListTestCase(unittest.TestCase):

    def setUp(self):
        if TEST_SETTINGS["version"] in ["2014", "2016", "2019", "365"]:
            version = Version.v2016
        else:
            version = Version.v2007

        authcookie = Office365(TEST_SETTINGS["server_url"], username=TEST_SETTINGS["username"], password=TEST_PASSWORD).GetCookies()
        self.site = Site(TEST_SETTINGS["site_url"], version=version, authcookie=authcookie)
        self.test_list = TEST_SETTINGS["test_list"]

    def tearDown(self):
        self.site._session.close()

    def test_a_create_list(self):
        print("Create List")
        self.site.AddList(self.test_list, description='Great List!', template_id='Custom List')
        self.assertTrue(self.test_list in [i['Title'] for i in self.site.get_list_collection()])

    def test_b_get_list_fields(self):
        print("Get Fields")
        self.list = self.site.List(TEST_SETTINGS["test_list"])
        self.assertIsNotNone(self.list.fields)

    def test_c_update_list(self):
        print("Update List")
        self.list = self.site.List(TEST_SETTINGS["test_list"])
        my_data = [{'Title': 'First Row!'},
                   {'Title': 'Another One!'},
                   {'Title': 'Thrid Row'}]
        self.list.UpdateListItems(data=my_data, kind='New')
        self.assertEqual(len(self.list.get_list_items(row_limit=5)), 3)

    def test_d_delete_row(self):
        print("Delete Row")
        self.list = self.site.List(TEST_SETTINGS["test_list"])
        my_data = [1]
        self.list.UpdateListItems(data=my_data, kind='Delete')
        self.assertEqual(len(self.list.get_list_items(row_limit=2)), 2)

    def test_e_get_view(self):
        print("Get View")
        self.list = self.site.List(TEST_SETTINGS["test_list"])
        self.assertEqual(len(self.list.GetListItems("All Items")), 2)

    def test_f_query_list(self):
        print('Test Query')
        self.list = self.site.List(TEST_SETTINGS["test_list"])
        query = {'Where': [('Eq', 'Title', 'Another One!')]}
        items = self.list.GetListItems(fields=['Title'], query=query)
        self.assertEqual(len(items), 1)

    def test_g_users(self):
        print("Test Users")
        self.list = self.site.List(TEST_SETTINGS["test_list"])
        self.assertIsNotNone(self.list.users)
        self.site.delete_list(TEST_SETTINGS["test_list"])



from shareplumfreightplusio import Site
from shareplumfreightplusio import Office365
from shareplumfreightplusio.site import Version
from .test_settings import TEST_SETTINGS, TEST_PASSWORD
import unittest

# Edit test_server.json file to setup SharePoint Test Server
# Use OS Enviroment Variable TEST_PASSWORD for SharePoint password
# export TEST_PASSWORD=********


class FolderTestCase(unittest.TestCase):

    def setUp(self):
        if TEST_SETTINGS["version"] in ["2014", "2016", "2019", "365"]:
            version = Version.v2016
        else:
            version = Version.v2007

        authcookie = Office365(TEST_SETTINGS["server_url"], username=TEST_SETTINGS["username"], password=TEST_PASSWORD).GetCookies()
        self.site = Site(TEST_SETTINGS["site_url"], version=version, authcookie=authcookie)

    def tearDown(self):
        self.site._session.close()

    def test_folder(self):
        print("Testing Folder")
        self.folder = self.site.Folder(TEST_SETTINGS["test_folder"])
        self.folder.upload_file("Hello", "new.txt")
        self.assertEqual(self.folder.get_file("new.txt"), b"Hello")
        self.folder.delete_file("new.txt")

        print("Testing Folder and files with apostrophes")
        self.folder = self.site.Folder(TEST_SETTINGS["test_folder_apostrophe"])
        self.folder.upload_file("Hello", "new'.txt")
        self.assertEqual(self.folder.get_file("new'.txt"), b"Hello")
        self.folder.delete_file("new'.txt")



import re

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from codecs import open

with open('shareplumfreightplusio/version.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

setup(
    name='shareplumfreightplusio',
    version=version,
    description='Python SharePoint FreightPlus Library',
    long_description=open('README.rst').read(),
    url='https://github.com/rcwethey/shareplumFreightplusIO',
    author='Ryan Wethey',
    author_email='rwethey@freightplus.io',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Office/Business',
    ],
    keywords=['SharePoint'],
    packages=['shareplumfreightplusio'],
    install_requires=['lxml', 'requests', 'requests-ntlm', 'requests-toolbelt'],
)



# This is a group of small functions
# used to work with a list of dictionaries
from typing import Any
from typing import Dict
from typing import List


def changes(new_cmp_dict, old_cmp_dict, id_column, columns):
    # type: (Dict, Dict, str, List[str]) -> List[Dict]
    """Return a list dict of the changes of the
       rows that exist in both dictionaries
       User must provide an ID column for old_cmp_dict
    """

    update_ldict = []
    same_keys = set(new_cmp_dict).intersection(set(old_cmp_dict))
    for same_key in same_keys:
        # Get the Union of the set of keys
        # for both dictionaries to account
        # for missing keys
        old_dict = old_cmp_dict[same_key]
        new_dict = new_cmp_dict[same_key]

        update_dict = {}
        for dict_key in columns:
            old_val = old_dict.get(dict_key, "NaN")
            new_val = new_dict.get(dict_key, "NaN")
            if old_val != new_val and new_val != "NaN":
                if id_column:
                    try:
                        update_dict[id_column] = old_dict[id_column]
                    except KeyError as e:
                        print("Input Dictionary 'old_cmp_dict' must have ID column: " + str(e))
                update_dict[dict_key] = new_val
        if update_dict:
            update_ldict.append(update_dict)
    return update_ldict


def unique(new_cmp_dict, old_cmp_dict):
    # type: (Dict, Dict) -> List
    """Return a list dict of
       the unique keys in new_cmp_dict
    """
    newkeys = set(new_cmp_dict)
    oldkeys = set(old_cmp_dict)
    unique = newkeys - oldkeys
    unique_ldict = []
    for key in unique:
        unique_ldict.append(new_cmp_dict[key])
    return unique_ldict


def full_dict(ldict, keys):
    # type: (Dict, Any) -> Dict
    """Return Comparison Dictionaries
       from list dict on keys
       keys: a list of keys that when
       combined make the row in the list unique
    """
    if type(keys) == str:
        keys = [keys]
    else:
        keys = keys

    cmp_dict = {}
    for line in ldict:
        index = []  # type: List[str]
        for key in keys:
            index.append(str(line.get(key, "")))
        index_str = "-".join(index)  # type: str
        cmp_dict[index_str] = line

    return cmp_dict



from .request_helper import get, post
import json


class _Folder():
    def __init__(self, session, folder_name, url, timeout=None):
        self._session = session
        self.folder_name = folder_name
        self._escaped_folder_name = self._escape_name(self.folder_name)
        self.site_url = url
        self.timeout = timeout

        self.info = self._create_folder()
        self._escaped_relative_url = self._escape_name(self.info['d']['ServerRelativeUrl'])

    @property
    def contextinfo(self):
        response = post(self._session, self.site_url + "/_api/contextinfo")
        data = response.json()
        return data

    def _escape_name(self, name):
        return name.replace("'", "''")

    def _create_folder(self):
        update_data = {}
        update_data['__metadata'] = {'type': 'SP.Folder'}
        update_data['ServerRelativeUrl'] = self.folder_name
        body = json.dumps(update_data)

        url = self.site_url + f"/_api/web/folders"

        headers = {'Accept': 'application/json;odata=verbose',
                   'Content-Type': 'application/json;odata=verbose',
                   'X-RequestDigest': self.contextinfo['FormDigestValue']}

        response = post(self._session, url=url, headers=headers, data=body, timeout=self.timeout)

        return response.json()

    def delete_folder(self, relative_url):
        if relative_url == self.folder_name:
            url = self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self._escaped_folder_name}')"

            headers = {'Accept': 'application/json;odata=verbose',
                       'If-Match': '*',
                       'X-HTTP-Method': 'DELETE',
                       'Content-Type': 'application/json;odata=verbose',
                       'X-RequestDigest': self.contextinfo['FormDigestValue']}

            post(self._session, url=url, headers=headers)
        else:
            print('You must pass the relative folder url to delete a folder')

    def delete_file(self, file_name):
        escaped_file_name = self._escape_name(file_name)
        url = self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self._escaped_relative_url}/{escaped_file_name}')"

        headers = {'Accept': 'application/json;odata=verbose',
                   'If-Match': '*',
                   'X-HTTP-Method': 'DELETE',
                   'Content-Type': 'application/json;odata=verbose',
                   'X-RequestDigest': self.contextinfo['FormDigestValue']}

        post(self._session, url=url, headers=headers)

    @property
    def items(self):
        response = get(self._session, self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self._escaped_folder_name}')/ListItemAllFields")
        return response.json()

    @property
    def files(self):
        response = get(self._session, self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self._escaped_folder_name}')/files")
        return response.json()['value']

    @property
    def folders(self):
        response = get(self._session, self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self._escaped_folder_name}')/folders")
        return [entry['Name'] for entry in response.json()['value']]

    def upload_file(self, content, file_name):
        escaped_file_name = self._escape_name(file_name)
        url = self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self._escaped_folder_name}')/Files/add(url='{escaped_file_name}',overwrite=true)"
        headers = {'X-RequestDigest': self.contextinfo['FormDigestValue']}

        post(self._session, url=url, headers=headers, data=content, timeout=self.timeout)

    def check_out(self, file_name):
        escaped_file_name = self._escape_name(file_name)
        url = self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self._escaped_relative_url}/{escaped_file_name}')/CheckOut()"
        headers = {'X-RequestDigest': self.contextinfo['FormDigestValue']}

        post(self._session, url=url, headers=headers)

    def check_in(self, file_name, comment):
        escaped_file_name = self._escape_name(file_name)
        url = self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self._escaped_relative_url}/{escaped_file_name}')/CheckIn(comment='{comment}',checkintype=0)"
        headers = {'X-RequestDigest': self.contextinfo['FormDigestValue']}
        post(self._session, url=url, headers=headers)

    def get_file(self, file_name):
        escaped_file_name = self._escape_name(file_name)
        response = get(self._session, self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self._escaped_relative_url}/{escaped_file_name}')/$value")
        return response.content
    
    def get_file_properties(self, file_name):
        file_properties= get(self._session, self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self.info['d']['ServerRelativeUrl']}/{file_name}')?/$expand=ListItemAllFields")
        return file_properties.json()
